# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

# Address families
AF_UNSPEC = 0
AF_INET = 2
AF_IPX = 6
AF_APPLETALK = 16
AF_NETBIOS = 17
AF_INET6 = 23
AF_IRDA = 26
AF_BTH = 32

# Socket types
SOCK_STREAM = 1
SOCK_DGRAM = 2
SOCK_RAW = 3
SOCK_RDM = 4
SOCK_SEQPACKET = 5

# Protocol types
IPPROTO_ICMP = 1
IPPROTO_IGMP = 2
BTHPROTO_RFCOMM = 3
IPPROTO_TCP = 6
IPPROTO_UDP = 17
IPPROTO_ICMPV6 = 58
IPPROTO_RM = 113

WSA_FLAG_OVERLAPPED = 1
WSA_FLAG_ACCESS_SYSTEM_SECURITY = 0x40
WSA_FLAG_NO_HANDLE_INHERIT = 0x80

HOST_NOT_FOUND = 11001
WSAENOTSOCK = 10038

MSG_PEEK = 0x2


def get_flag_defines(flags, prefix=''):
    defs = []
    for k, v in globals().items():
        if not isinstance(v, int):
            continue
        if v & flags:
            if prefix and k.startswith(prefix):
                defs.append(k)
    return defs


def get_define(define, prefix=''):
    for k, v in globals().items():
        if not isinstance(v, int) or v != define:
            continue
        if prefix:
            if k.startswith(prefix):
                return k
        else:
            return k


def get_addr_family(define):
    return get_define(define, prefix='AF_')


def get_sock_type(define):
    return get_define(define, prefix='SOCK_')


def get_proto_type(define):
    return get_define(define, prefix='IPPROTO_')



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from socket import inet_ntoa, ntohs, htons, ntohl, htonl, inet_aton

import speakeasy.winenv.arch as _arch

import speakeasy.winenv.defs.windows.windows as windefs
import speakeasy.winenv.defs.winsock.winsock as winsock
import speakeasy.winenv.defs.winsock.ws2_32 as wstypes

from .. import api

INADDR_NONE = 0xFFFFFFFF
HOST_NOT_FOUND = 11001
WSAENOTSOCK = 10038


class Ws2_32(api.ApiHandler):
    """
    Implements winsock functions from ws2_32.dll
    """

    name = 'ws2_32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Ws2_32, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.addr_bufs = {}
        self.last_error = 0
        self.win = None
        self.netman = emu.get_network_manager()
        self.wstypes = wstypes

        super(Ws2_32, self).__get_hook_attrs__(self)

    @apihook('WSAStartup', argc=2, conv=_arch.CALL_CONV_STDCALL, ordinal=115)
    def WSAStartup(self, emu, argv, ctx={}):
        """
        int WSAStartup(
          WORD      wVersionRequired,
          LPWSADATA lpWSAData
        );
        """
        ver, lpWSAData = argv

        wsa = self.wstypes.WSAData(emu.get_ptr_size())
        data = self.mem_cast(wsa, lpWSAData)

        data.wVersion = 0x0101
        data.wHighVersion = 0x0202
        data.iMaxSockets = 0x1000
        data.iMaxUdpDg = 0x1000

        self.mem_write(lpWSAData, self.get_bytes(data))

        rv = windefs.ERROR_SUCCESS
        return rv

    @apihook('WSACleanup', argc=0, ordinal=116)
    def WSACleanup(self, emu, argv, ctx={}):
        """
        int WSACleanup();
        """

        return 0

    @apihook('WSASocket', argc=6)
    def WSASocket(self, emu, argv, ctx={}):
        """
        SOCKET WSAAPI WSASocket(
          int                 af,
          int                 type,
          int                 protocol,
          LPWSAPROTOCOL_INFO  lpProtocolInfo,
          GROUP               g,
          DWORD               dwFlags
        );
        """
        af, typ, protocol, lpProtocolInfo, g, dwFlags = argv

        fam_str = winsock.get_addr_family(af)
        sock_str = winsock.get_sock_type(typ)

        sock = self.netman.new_socket(fam_str, sock_str, protocol, dwFlags)

        fd = sock.get_fd()

        argv[0] = fam_str
        argv[1] = sock_str

        return fd

    @apihook('socket', argc=3, conv=_arch.CALL_CONV_STDCALL, ordinal=23)
    def socket(self, emu, argv, ctx={}):
        """
        SOCKET WSAAPI socket(
          int af,
          int type,
          int protocol
        );
        """
        af, typ, protocol = argv

        fam_str = winsock.get_addr_family(af)
        sock_str = winsock.get_sock_type(typ)

        sock = self.netman.new_socket(fam_str, sock_str, protocol, 0)

        fd = sock.get_fd()

        argv[0] = fam_str
        argv[1] = sock_str

        return fd

    @apihook('inet_addr', argc=1, ordinal=11)
    def inet_addr(self, emu, argv, ctx={}):
        """
    unsigned long inet_addr(
      _In_ const char *cp
    );
        """
        a, = argv

        if a:
            a = self.read_mem_string(a, 1)
            argv[0] = a
            try:
                rv = inet_aton(a)
                rv = int.from_bytes(rv, 'little')
            except OSError:
                rv = INADDR_NONE

        return rv

    @apihook('htons', argc=1, conv=_arch.CALL_CONV_STDCALL, ordinal=9)
    def htons(self, emu, argv, ctx={}):
        """
        u_short htons(
          u_short hostshort
        );
        """
        hostshort, = argv

        netshort = htons(hostshort)

        return netshort

    @apihook('ntohs', argc=1, ordinal=15)
    def ntohs(self, emu, argv, ctx={}):
        """
        u_short ntohs(
            u_short netshort
        );
        """
        netshort, = argv

        return ntohs(netshort)

    @apihook('ntohl', argc=1, ordinal=14)
    def ntohl(self, emu, argv, ctx={}):
        """
        u_long ntohl(
            u_long netlong
        );
        """
        netlong, = argv

        return ntohl(netlong)

    @apihook('setsockopt', argc=5, ordinal=21)
    def setsockopt(self, emu, argv, ctx={}):
        """
        int setsockopt(
          SOCKET     s,
          int        level,
          int        optname,
          const char *optval,
          int        optlen
        );
        """
        s, level, optname, optval, optlen = argv

        return 0

    @apihook('WSASetLastError', argc=1, ordinal=112)
    def WSASetLastError(self, emu, argv, ctx={}):
        """
        void WSASetLastError(
            int iError
        );
        """
        iError, = argv

        self.last_error = iError
        return

    @apihook('gethostname', argc=2, ordinal=57)
    def gethostname(self, emu, argv, ctx={}):
        """
        int gethostname(
            char *name,
            int  namelen
        );
        """
        name, namelen, = argv
        rv = -1

        host = emu.get_hostname()
        if name and host:
            host += '\x00'
            if namelen > len(host):
                out = host.encode('utf-8')
                self.mem_write(name, out)
                rv = 0
        return rv

    @apihook('gethostbyname', argc=1, conv=_arch.CALL_CONV_STDCALL, ordinal=52)
    def gethostbyname(self, emu, argv, ctx={}):
        """
        struct hostent * gethostbyname(const char FAR * name);
        """
        name, = argv

        name = self.read_mem_string(name, 1)
        ptr_hostent = 0

        hostent = 0
        ip = self.netman.name_lookup(name)

        if ip:
            hostent = self.wstypes.hostent(emu.get_ptr_size())
            buflen = len(name)
            buflen += 8
            buflen += len(ip)
            ptr_hostent = self.mem_alloc(buflen, tag='api.struct.HOSTENT')
            hostent.h_name = argv[0]
            hostent.h_length = len(ip)
            ip_bytes = inet_aton(ip)
            ptr_h_addr = self.mem_alloc(len(ip_bytes)*3, tag='api.struct.HOSTENT.h_addr')

            ptr = (ptr_h_addr + self.get_ptr_size())
            self.mem_write(ptr_h_addr, ptr.to_bytes(self.get_ptr_size(), 'little'))
            self.mem_write(ptr, ip_bytes)
            hostent.h_addr_list = ptr_h_addr

            # Write the hostent struct
            self.mem_write(ptr_hostent, self.get_bytes(hostent))
        else:
            ip = ''

        argv[0] = name
        self.log_dns(name, ip)

        return ptr_hostent

    @apihook('connect', argc=3, conv=_arch.CALL_CONV_STDCALL, ordinal=4)
    def connect(self, emu, argv, ctx={}):
        """
        int WSAAPI connect(
          SOCKET         s,
          const sockaddr *name,
          int            namelen
        );
        """

        s, pname, namelen = argv

        rv = windefs.ERROR_SUCCESS

        sockaddr = self.wstypes.sockaddr_in(emu.get_ptr_size())
        sa = self.mem_cast(sockaddr, pname)

        raddr = inet_ntoa(sa.sin_addr.to_bytes(4, 'little'))
        rport = ntohs(sa.sin_port)

        socket = self.netman.get_socket(s)
        if not socket:
            return 0xFFFFFFFF
        stype = socket.get_type()
        proto = 'unknown'
        if stype == 'SOCK_STREAM':
            proto = 'tcp'
        elif stype == 'SOCK_DGRAM':
            proto = 'udp'
        elif stype == 'SOCK_RAW':
            proto = 'raw'

        socket.set_connection_info(raddr, rport)

        self.log_network(raddr, rport, typ='connect', proto=proto, method='winsock.connect')

        argv[1] = '%s:%d' % (raddr, rport)

        return rv

    @apihook('bind', argc=3, ordinal=2)
    def bind(self, emu, argv, ctx={}):
        """
        int bind(
            SOCKET         s,
            const sockaddr *addr,
            int            namelen
        );
        """
        s, pname, namelen = argv
        rv = windefs.ERROR_SUCCESS

        sockaddr = self.wstypes.sockaddr_in(emu.get_ptr_size())
        sa = self.mem_cast(sockaddr, pname)
        raddr = inet_ntoa(sa.sin_addr.to_bytes(4, 'little'))
        rport = ntohs(sa.sin_port)

        socket = self.netman.get_socket(s)
        stype = socket.get_type()
        proto = 'unknown'
        if stype == 'SOCK_STREAM':
            proto = 'tcp'
        elif stype == 'SOCK_DGRAM':
            proto = 'udp'
        elif stype == 'SOCK_RAW':
            proto = 'raw'

        socket.set_connection_info(raddr, rport)
        self.log_network(raddr, rport, typ='bind', proto=proto, method='winsock.bind')

        argv[1] = '%s:%d' % (raddr, rport)

        return rv

    @apihook('listen', argc=2, ordinal=13)
    def listen(self, emu, argv, ctx={}):
        """
        int WSAAPI listen(
            SOCKET s,
            int    backlog
        );
        """
        s, backlog = argv
        rv = windefs.ERROR_SUCCESS

        return rv

    @apihook('select', argc=5, ordinal=18)
    def select(self, emu, argv, ctx={}):
        """
        int WSAAPI select(
            int           nfds,
            fd_set        *readfds,
            fd_set        *writefds,
            fd_set        *exceptfds,
            const timeval *timeout
        );
        """
        nfds, readfds, writefds, exceptfds, timeout = argv
        fd_count = 0

        if readfds:
            fds = self.mem_read(readfds, 4)
            fds = int.from_bytes(fds, 'little')
            fd_count += fds

        if writefds:
            fds = self.mem_read(writefds, 4)
            fds = int.from_bytes(fds, 'little')
            fd_count += fds

        if exceptfds:
            fds = self.mem_read(exceptfds, 4)
            fds = int.from_bytes(fds, 'little')
            fd_count += fds

        return fd_count

    @apihook('accept', argc=3, ordinal=1)
    def accept(self, emu, argv, ctx={}):
        """
        SOCKET WSAAPI accept(
            SOCKET   s,
            sockaddr *addr,
            int      *addrlen
        );
        """
        s, addr, addrlen = argv

        socket = self.netman.get_socket(s)
        stype = socket.get_type()
        proto = 'unknown'
        if stype == 'SOCK_STREAM':
            proto = 'tcp'
        elif stype == 'SOCK_DGRAM':
            proto = 'udp'
        elif stype == 'SOCK_RAW':
            proto = 'raw'

        new_sock = self.netman.new_socket(socket.family, socket.type, socket.protocol, 0)
        aip = self.netman.name_lookup('default')
        if not aip:
            aip = '127.0.0.1'

        port = socket.connected_port
        nip = inet_aton(aip)
        nip = int.from_bytes(nip, 'little')

        new_sock.set_connection_info(aip, port)

        self.log_network(aip, port, typ='accept', proto=proto, method='winsock.accept')

        if addr:
            sockaddr = self.wstypes.sockaddr_in(emu.get_ptr_size())
            sockaddr = self.mem_cast(sockaddr, addr)
            sockaddr.sin_addr = nip
            sockaddr.sin_port = port
            self.mem_write(addr, sockaddr.get_bytes())

        return new_sock.get_fd()

    @apihook('inet_ntoa', argc=1, ordinal=12)
    def inet_ntoa(self, emu, argv, ctx={}):
        """
        char FAR* inet_ntoa(struct in_addr in);
        """
        in_addr, = argv

        raddr = inet_ntoa(in_addr.to_bytes(4, 'little'))
        rv = self.addr_bufs.get(raddr)
        if not rv:
            buf = self.mem_alloc(len(raddr), tag='api.ws2_32.inet_ntoa.%s' % (raddr))
            self.mem_write(buf, raddr.encode('utf-8'))
            self.addr_bufs.update({raddr: buf})
        return buf

    @apihook('htonl', argc=1, ordinal=8)
    def htonl(self, emu, argv, ctx={}):
        """
        uint32_t htonl(uint32_t hostlong);
        """
        hostlong, = argv
        return htonl(hostlong)

    @apihook('__WSAFDIsSet', argc=2, ordinal=151)
    def __WSAFDIsSet(self, emu, argv, ctx={}):
        """
        int __WSAFDIsSet(
            SOCKET ,
            fd_set *
        );
        """
        sock, fd_set = argv
        return 1

    @apihook('shutdown', argc=2, ordinal=22)
    def shutdown(self, emu, argv, ctx={}):
        """
        int shutdown(
            SOCKET s,
            int    how
        );
        """
        return 0

    @apihook('recv', argc=4, ordinal=16)
    def recv(self, emu, argv, ctx={}):
        """
        int recv(
          SOCKET s,
          char   *buf,
          int    len,
          int    flags
        );
        """

        s, buf, blen, flags = argv
        rv = 0

        peek = flags & winsock.MSG_PEEK

        sock = self.netman.get_socket(s)
        data = sock.get_recv_data(blen, peek)
        rv = len(data)

        self.mem_write(buf, data)

        stype = sock.get_type()
        proto = 'unknown'
        if stype == 'SOCK_STREAM':
            proto = 'tcp'
        elif stype == 'SOCK_DGRAM':
            proto = 'udp'
        elif stype == 'SOCK_RAW':
            proto = 'raw'

        raddr, rport = sock.get_connection_info()
        self.log_network(raddr, rport, typ='data_in', proto=proto, method='winsock.recv',
                         data=data)

        return rv

    @apihook('send', argc=4, ordinal=19)
    def send(self, emu, argv, ctx={}):
        """
        int WSAAPI send(
          SOCKET     s,
          const char *buf,
          int        len,
          int        flags
        );
        """
        s, buf, blen, flags = argv
        data = b''

        socket = self.netman.get_socket(s)
        stype = socket.get_type()
        proto = 'unknown'
        if stype == 'SOCK_STREAM':
            proto = 'tcp'
        elif stype == 'SOCK_DGRAM':
            proto = 'udp'
        elif stype == 'SOCK_RAW':
            proto = 'raw'

        if buf:
            data = self.mem_read(buf, blen)
        raddr, rport = socket.get_connection_info()

        self.log_network(raddr, rport, typ='data_out', proto=proto, method='winsock.send',
                         data=data)

        return len(data)

    @apihook('closesocket', argc=1, ordinal=3)
    def closesocket(self, emu, argv, ctx={}):
        """
        int closesocket(
          IN SOCKET s
        );
        """
        s, = argv

        rv = 0

        socket = self.netman.get_socket(s)
        if not socket:
            # This isnt a valid socket, return invalid
            rv = winsock.WSAENOTSOCK
        else:
            self.netman.close_socket(s)

        return rv

    @apihook('ioctlsocket', argc=3, ordinal=10)
    def ioctlsocket(self, emu, argv, ctx={}):
        """
        int ioctlsocket(
            SOCKET s,
            long   cmd,
            u_long *argp
        );
        """
        s, cmd, argp = argv
        rv = winsock.WSAENOTSOCK

        socket = self.netman.get_socket(s)
        if socket:
            rv = 0

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.


import unittest

import util


class TestFileAccess(unittest.TestCase):

    def setUp(self):
        self.config = util.get_config()
        self.report = None

    def _get_api_calls(self, ep, api_name):
        return [api for api in ep['apis'] if api['api_name'] == api_name]

    def _test_file_access(self, fpath):
        data = util.get_test_bin_data(fpath)
        self.report = util.run_test(self.config, data)
        eps = self.report['entry_points']

        driver_entry = eps[0]

        create_file = self._get_api_calls(driver_entry, 'ntdll.NtCreateFile')
        self.assertEqual(1, len(create_file))
        create_file = create_file[0]
        self.assertEqual(create_file['args'][3], '\\??\\c:\\myfile.txt')

        read_file = self._get_api_calls(driver_entry, 'ntdll.NtReadFile')
        self.assertEqual(1, len(read_file))

        printf = self._get_api_calls(driver_entry,
                                     'api-ms-win-crt-stdio-l1-1-0.__stdio_common_vfprintf')
        self.assertEqual(5, len(printf))
        printf = printf[-1]

        self.assertIn('File contained:', printf['args'][2])

    def test_file_access_x86(self):
        self._test_file_access('file_access_test_x86.exe.xz')

    def test_file_access_x64(self):
        self._test_file_access('file_access_test_x64.exe.xz')


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import uuid
import speakeasy.winenv.arch as e_arch

import speakeasy.winenv.defs.nt.ddk as ddk
import speakeasy.winenv.defs.wdf as wdf
import speakeasy.winenv.defs.usb as usbdefs

import speakeasy.winenv.defs.nt.ntoskrnl as ntos

from .. import api


class WdfDriver(object):
    def __init__(self):
        self.reg_path = None
        self.typed_context_worker = None
        self.queues = {}
        self.driver_object_addr = None
        self.driver_object = None


class WdfDevice(object):
    def __init__(self):
        self.device_object_addr = None
        self.device_object = None


class WdfUsbDevice(object):
    def __init__(self):
        self.num_interfaces = 0
        self.config_desc = None


class WdfUsbInterface(object):
    def __init__(self):
        self.config_desc = 0
        self.iface_index = 0
        self.setting_index = 0


class WdfUsbPipe(object):
    def __init__(self):
        self.interface = None
        self.index = 0


class Wdfldr(api.ApiHandler):
    """
    Implements the Windows Driver Framework (WDK)
    """

    name = 'wdfldr'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Wdfldr, self).__init__(emu)

        self.funcs = {}
        self.curr_handle = 4
        self.pnp_device = None
        self.data = {}
        self.emu = emu
        self.wdf_drivers = {}
        self.wdf_devices = {}
        self.usb_devices = {}
        self.usb_pipes = {}
        self.usb_interfaces = {}
        self.handles = {}
        self.types = wdf
        self.func_table = self.types.WDFFUNCTIONS(emu.get_ptr_size())
        self.func_table_ptr = None
        self.component_globals = None
        super(Wdfldr, self).__get_hook_attrs__(self)

    def get_handle(self):
        self.curr_handle += 4
        return self.curr_handle

    def set_func_table(self, emu):

        addr = emu.add_callback(Wdfldr.name, self.WdfDriverCreate.__apihook__[0])
        self.func_table.pfnWdfDriverCreate = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDeviceInitSetPnpPowerEventCallbacks.__apihook__[0])
        self.func_table.pfnWdfDeviceInitSetPnpPowerEventCallbacks = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDeviceInitSetRequestAttributes.__apihook__[0])
        self.func_table.pfnWdfDeviceInitSetRequestAttributes = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDeviceInitSetFileObjectConfig.__apihook__[0])
        self.func_table.pfnWdfDeviceInitSetFileObjectConfig = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDeviceInitSetIoType.__apihook__[0])
        self.func_table.pfnWdfDeviceInitSetIoType = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDeviceCreate.__apihook__[0])
        self.func_table.pfnWdfDeviceCreate = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfObjectGetTypedContextWorker.__apihook__[0])
        self.func_table.pfnWdfObjectGetTypedContextWorker = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDriverOpenParametersRegistryKey.__apihook__[0])
        self.func_table.pfnWdfDriverOpenParametersRegistryKey = addr

        addr = emu.add_callback(Wdfldr.name, self.WdfRegistryQueryULong.__apihook__[0])
        self.func_table.pfnWdfRegistryQueryULong = addr

        addr = emu.add_callback(Wdfldr.name, self.WdfRegistryClose.__apihook__[0])
        self.func_table.pfnWdfRegistryClose = addr

        addr = emu.add_callback(Wdfldr.name, self.WdfDeviceSetPnpCapabilities.__apihook__[0])
        self.func_table.pfnWdfDeviceSetPnpCapabilities = addr

        addr = emu.add_callback(Wdfldr.name, self.WdfIoQueueCreate.__apihook__[0])
        self.func_table.pfnWdfIoQueueCreate = addr

        addr = emu.add_callback(Wdfldr.name, self.WdfIoQueueReadyNotify.__apihook__[0])
        self.func_table.pfnWdfIoQueueReadyNotify = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDeviceCreateDeviceInterface.__apihook__[0])
        self.func_table.pfnWdfDeviceCreateDeviceInterface = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDeviceWdmGetAttachedDevice.__apihook__[0])
        self.func_table.pfnWdfDeviceWdmGetAttachedDevice = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfDeviceWdmGetDeviceObject.__apihook__[0])
        self.func_table.pfnWdfDeviceWdmGetDeviceObject = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbTargetDeviceCreateWithParameters.__apihook__[0])
        self.func_table.pfnWdfUsbTargetDeviceCreateWithParameters = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbTargetDeviceGetDeviceDescriptor.__apihook__[0])
        self.func_table.pfnWdfUsbTargetDeviceGetDeviceDescriptor = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbTargetDeviceRetrieveConfigDescriptor.__apihook__[0])
        self.func_table.pfnWdfUsbTargetDeviceRetrieveConfigDescriptor = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfMemoryCreate.__apihook__[0])
        self.func_table.pfnWdfMemoryCreate = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbTargetDeviceSelectConfig.__apihook__[0])
        self.func_table.pfnWdfUsbTargetDeviceSelectConfig = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbTargetDeviceGetNumInterfaces.__apihook__[0])
        self.func_table.pfnWdfUsbTargetDeviceGetNumInterfaces = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbTargetDeviceRetrieveInformation.__apihook__[0])
        self.func_table.pfnWdfUsbTargetDeviceRetrieveInformation = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbInterfaceGetNumSettings.__apihook__[0])
        self.func_table.pfnWdfUsbInterfaceGetNumSettings = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbInterfaceSelectSetting.__apihook__[0])
        self.func_table.pfnWdfUsbInterfaceSelectSetting = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbInterfaceGetNumConfiguredPipes.__apihook__[0])
        self.func_table.pfnWdfUsbInterfaceGetNumConfiguredPipes = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbInterfaceGetConfiguredPipe.__apihook__[0])
        self.func_table.pfnWdfUsbInterfaceGetConfiguredPipe = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbTargetPipeGetInformation.__apihook__[0])
        self.func_table.pfnWdfUsbTargetPipeGetInformation = addr

        addr = emu.add_callback(Wdfldr.name,
                                self.WdfUsbInterfaceGetInterfaceNumber.__apihook__[0])
        self.func_table.pfnWdfUsbInterfaceGetInterfaceNumber = addr

        self.mem_write(self.func_table_ptr, self.func_table.get_bytes())

    def parse_usb_config(self, data):

        interfaces = []

        # Get the USB config descriptor
        cd = usbdefs.USB_CONFIGURATION_DESCRIPTOR().cast(data)
        ifaces = cd.bNumInterfaces
        data = data[cd.bLength:]

        for i in range(ifaces):
            endpoints = []
            _id = usbdefs.USB_INTERFACE_DESCRIPTOR().cast(data)

            data = data[_id.bLength:]
            for j in range(_id.bNumEndpoints):
                ep = usbdefs.USB_ENDPOINT_DESCRIPTOR().cast(data)
                data = data[ep.bLength:]
                endpoints.append(ep)

            interfaces.append([_id, endpoints])

        return interfaces

    @apihook('WdfVersionBind', argc=4)
    def WdfVersionBind(self, emu, argv, ctx={}):
        '''
        NTSTATUS
        WdfVersionBind(
        __in PDRIVER_OBJECT DriverObject,
        __in PUNICODE_STRING RegistryPath,
        __inout PWDF_BIND_INFO BindInfo,
        __out PWDF_COMPONENT_GLOBALS* ComponentGlobals
        );
        '''
        rv = ddk.STATUS_SUCCESS
        drv, reg_path, BindInfo, comp_globals = argv

        wbi = self.types.WDF_BIND_INFO(emu.get_ptr_size())
        wbi = self.mem_cast(wbi, BindInfo)

        if not self.func_table_ptr:
            size = self.func_table.sizeof()
            self.func_table_ptr = self.mem_alloc(size, tag='api.struct.WDFFUNCTIONS')

            self.mem_write(wbi.FuncTable, (self.func_table_ptr).to_bytes(emu.get_ptr_size(),
                                                                         'little'))

        if not self.component_globals:
            components = self.types.WDF_COMPONENT_GLOBALS(emu.get_ptr_size())
            self.component_globals = self.mem_alloc(components.sizeof(),
                                                    tag='api.struct.WDF_COMPONENT_GLOBALS')
            self.mem_write(comp_globals, (self.component_globals).to_bytes(emu.get_ptr_size(),
                                                                           'little'))

        self.set_func_table(emu)

        # For now, just leave the handle open so we can reference it later
        return rv

    @apihook('WdfDriverCreate', argc=6)
    def WdfDriverCreate(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfDriverCreate(
          PWDF_DRIVER_GLOBALS DriverGlobals,
          PDRIVER_OBJECT         DriverObject,
          PCUNICODE_STRING       RegistryPath,
          PWDF_OBJECT_ATTRIBUTES DriverAttributes,
          PWDF_DRIVER_CONFIG     DriverConfig,
          WDFDRIVER              *Driver
        );
        '''
        DriverGlobals, DriverObject, RegistryPath, DriverAttributes, DriverConfig, Driver = argv

        driver = WdfDriver()
        driver.reg_path = self.read_unicode_string(RegistryPath)
        driver.driver_object_addr = DriverObject
        driver.driver_object = self.mem_cast(ntos.DRIVER_OBJECT(emu.get_ptr_size()),
                                             DriverObject)

        self.wdf_drivers.update({DriverGlobals: driver})

        if DriverConfig:
            config = self.mem_cast(self.types.WDF_DRIVER_CONFIG(emu.get_ptr_size()),  # noqa
                                   DriverConfig)
        rv = 0

        return rv

    @apihook('WdfDeviceInitSetPnpPowerEventCallbacks', argc=3)
    def WdfDeviceInitSetPnpPowerEventCallbacks(self, emu, argv, ctx={}):
        '''
        void WdfDeviceInitSetPnpPowerEventCallbacks(
          PWDFDEVICE_INIT               DeviceInit,
          PWDF_PNPPOWER_EVENT_CALLBACKS PnpPowerEventCallbacks
        );
        '''
        DriverGlobals, DeviceInit, PnpPowerEventCallbacks = argv

        return

    @apihook('WdfDeviceInitSetRequestAttributes', argc=3)
    def WdfDeviceInitSetRequestAttributes(self, emu, argv, ctx={}):
        '''
        void WdfDeviceInitSetRequestAttributes(
          PWDFDEVICE_INIT        DeviceInit,
          PWDF_OBJECT_ATTRIBUTES RequestAttributes
        );
        '''
        DriverGlobals, DeviceInit, RequestAttributes = argv

        return

    @apihook('WdfDeviceInitSetFileObjectConfig', argc=4)
    def WdfDeviceInitSetFileObjectConfig(self, emu, argv, ctx={}):
        '''
        void WdfDeviceInitSetFileObjectConfig(
          PWDFDEVICE_INIT        DeviceInit,
          PWDF_FILEOBJECT_CONFIG FileObjectConfig,
          PWDF_OBJECT_ATTRIBUTES FileObjectAttributes
        );
        '''
        DriverGlobals, DeviceInit, FileObjectConfig, FileObjectAttributes = argv

        return

    @apihook('WdfDeviceInitSetIoType', argc=3)
    def WdfDeviceInitSetIoType(self, emu, argv, ctx={}):
        '''
        void WdfDeviceInitSetIoType(
          PWDFDEVICE_INIT    DeviceInit,
          WDF_DEVICE_IO_TYPE IoType
        );
        '''
        DriverGlobals, DeviceInit, IoType = argv

        return

    @apihook('WdfDeviceCreate', argc=4)
    def WdfDeviceCreate(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfDeviceCreate(
          PWDFDEVICE_INIT        *DeviceInit,
          PWDF_OBJECT_ATTRIBUTES DeviceAttributes,
          WDFDEVICE              *Device
        );
        '''
        DriverGlobals, DeviceInit, DeviceAttributes, Device = argv
        rv = ddk.STATUS_SUCCESS

        if Device:
            handle = self.get_handle()
            dev = WdfDevice()
            self.wdf_devices.update({handle: dev})
            self.mem_write(Device, (handle).to_bytes(emu.get_ptr_size(), 'little'))

            do = ntos.DEVICE_OBJECT(emu.get_ptr_size())
            dev.device_object_addr = self.mem_alloc(do.sizeof(), tag='api.struct.DEVICE_OBJECT')
            dev.device_object = do

            driver = self.wdf_drivers.get(DriverGlobals)
            if driver:
                dev.device_object.DriverObject = driver.driver_object_addr
                self.mem_write(dev.device_object_addr, dev.device_object.get_bytes())

        return rv

    @apihook('WdfObjectGetTypedContextWorker', argc=3, conv=e_arch.CALL_CONV_FASTCALL)
    def WdfObjectGetTypedContextWorker(self, emu, argv, ctx={}):
        '''
        PVOID WdfObjectGetTypedContextWorker(
          WDFOBJECT                      Handle,
          PCWDF_OBJECT_CONTEXT_TYPE_INFO TypeInfo
        );
        '''
        DriverGlobals, Handle, TypeInfo = argv

        driver = self.wdf_drivers.get(DriverGlobals)

        if not driver.typed_context_worker:
            size = self.types.WDF_COMPONENT_GLOBALS(emu.get_ptr_size()).sizeof()
            driver.typed_context_worker = self.mem_alloc(size,
                                                         tag='api.struct.WDF_TYPED_CONTEXT_WORKER')
        rv = driver.typed_context_worker

        return rv

    @apihook('WdfDriverOpenParametersRegistryKey', argc=5)
    def WdfDriverOpenParametersRegistryKey(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfDriverOpenParametersRegistryKey(
          WDFDRIVER              Driver,
          ACCESS_MASK            DesiredAccess,
          PWDF_OBJECT_ATTRIBUTES KeyAttributes,
          WDFKEY                 *Key
        );
        '''
        DriverGlobals, Driver, DesiredAccess, KeyAttributes, pKey = argv

        rv = ddk.STATUS_OBJECT_NAME_NOT_FOUND

        driver = self.wdf_drivers.get(DriverGlobals)
        hnd = emu.reg_open_key(driver.reg_path + '\\Parameters')
        if hnd:
            rv = ddk.STATUS_SUCCESS

        if pKey:
            self.mem_write(pKey, (hnd).to_bytes(emu.get_ptr_size(), 'little'))

        return rv

    @apihook('WdfRegistryQueryULong', argc=4)
    def WdfRegistryQueryULong(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfRegistryQueryULong(
          WDFKEY           Key,
          PCUNICODE_STRING ValueName,
          PULONG           Value
        );
        '''
        DriverGlobals, Key, ValueName, Value = argv

        rv = ddk.STATUS_OBJECT_NAME_NOT_FOUND
        wkey = emu.reg_get_key(Key)
        if wkey:

            val_name = self.read_unicode_string(ValueName)
            argv[2] = val_name
            value = wkey.get_value(val_name)
            if value:
                ulong = value.get_data()
                self.mem_write(Value, (ulong).to_bytes(4, 'little'))
                rv = ddk.STATUS_SUCCESS

        return rv

    @apihook('WdfRegistryClose', argc=2)
    def WdfRegistryClose(self, emu, argv, ctx={}):
        '''
        void WdfRegistryClose(
          WDFKEY Key
        );
        '''
        DriverGlobals, Key = argv
        return

    @apihook('WdfDeviceSetPnpCapabilities', argc=3)
    def WdfDeviceSetPnpCapabilities(self, emu, argv, ctx={}):
        '''
        void WdfDeviceSetPnpCapabilities(
          WDFDEVICE                    Device,
          PWDF_DEVICE_PNP_CAPABILITIES PnpCapabilities
        );
        '''
        DriverGlobals, Device, PnpCapabilities = argv
        return

    @apihook('WdfIoQueueReadyNotify', argc=4)
    def WdfIoQueueReadyNotify(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfIoQueueReadyNotify(
          WDFQUEUE               Queue,
          PFN_WDF_IO_QUEUE_STATE QueueReady,
          WDFCONTEXT             Context
        );
        '''
        DriverGlobals, Queue, QueueReady, Context = argv
        rv = ddk.STATUS_SUCCESS

        return rv

    @apihook('WdfDeviceCreateDeviceInterface', argc=4)
    def WdfDeviceCreateDeviceInterface(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfDeviceCreateDeviceInterface(
          WDFDEVICE        Device,
          const GUID       *InterfaceClassGUID,
          PCUNICODE_STRING ReferenceString
        );
        '''
        DriverGlobals, Device, InterfaceClassGUID, ReferenceString = argv
        rv = ddk.STATUS_SUCCESS

        if InterfaceClassGUID:
            guid = self.mem_read(InterfaceClassGUID, 16)
            guid = uuid.UUID(bytes_le=guid)
            argv[2] = str(guid)

        if ReferenceString:
            ref = self.read_unicode_string(ReferenceString)
            argv[3] = ref

        return rv

    @apihook('WdfIoQueueCreate', argc=5)
    def WdfIoQueueCreate(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfIoQueueCreate(
          WDFDEVICE              Device,
          PWDF_IO_QUEUE_CONFIG   Config,
          PWDF_OBJECT_ATTRIBUTES QueueAttributes,
          WDFQUEUE               *Queue
        );
        '''
        DriverGlobals, Device, Config, QueueAttributes, Queue = argv
        rv = ddk.STATUS_SUCCESS

        queue_config = self.types.WDF_IO_QUEUE_CONFIG(emu.get_ptr_size())
        queue_config = self.mem_cast(queue_config, Config)

        hnd = self.get_handle()
        driver = self.wdf_drivers.get(DriverGlobals)
        driver.queues.update({hnd: queue_config})

        if Queue:
            self.mem_write(Queue, (hnd).to_bytes(emu.get_ptr_size(), 'little'))

        return rv

    @apihook('WdfDeviceWdmGetAttachedDevice', argc=2)
    def WdfDeviceWdmGetAttachedDevice(self, emu, argv, ctx={}):
        '''
        PDEVICE_OBJECT WdfDeviceWdmGetAttachedDevice(
          WDFDEVICE Device
        );
        '''
        DriverGlobals, Device = argv

        if not self.pnp_device:
            do = ntos.DEVICE_OBJECT(emu.get_ptr_size())
            self.pnp_device = self.mem_alloc(do.sizeof(), tag='api.struct.DEVICE_OBJECT')
        rv = self.pnp_device

        return rv

    @apihook('WdfUsbTargetDeviceCreateWithParameters', argc=5)
    def WdfUsbTargetDeviceCreateWithParameters(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfUsbTargetDeviceCreateWithParameters(
          WDFDEVICE                     Device,
          PWDF_USB_DEVICE_CREATE_CONFIG Config,
          PWDF_OBJECT_ATTRIBUTES        Attributes,
          WDFUSBDEVICE                  *UsbDevice
        );
        '''
        DriverGlobals, Device, Config, Attributes, UsbDevice = argv

        rv = ddk.STATUS_SUCCESS
        handle = self.get_handle()
        usb = WdfUsbDevice()
        self.usb_devices.update({handle: usb})

        self.mem_write(UsbDevice, (handle).to_bytes(emu.get_ptr_size(), 'little'))

        return rv

    @apihook('WdfDeviceWdmGetDeviceObject', argc=2)
    def WdfDeviceWdmGetDeviceObject(self, emu, argv, ctx={}):
        '''
        PDEVICE_OBJECT WdfDeviceWdmGetDeviceObject(
          WDFDEVICE Device
        );
        '''
        DriverGlobals, Device = argv
        rv = 0

        dev = self.wdf_devices.get(Device)
        if dev:
            rv = dev.device_object_addr
        return rv

    @apihook('WdfUsbTargetDeviceGetDeviceDescriptor', argc=3)
    def WdfUsbTargetDeviceGetDeviceDescriptor(self, emu, argv, ctx={}):
        '''
        void WdfUsbTargetDeviceGetDeviceDescriptor(
          WDFUSBDEVICE           UsbDevice,
          PUSB_DEVICE_DESCRIPTOR UsbDeviceDescriptor
        );
        '''
        DriverGlobals, UsbDevice, UsbDeviceDescriptor = argv

        dev = self.usb_devices.get(UsbDevice)
        if dev:
            dd = usbdefs.USB_DEVICE_DESCRIPTOR(emu.get_ptr_size())
            self.mem_write(UsbDeviceDescriptor, dd.get_bytes())
        return

    @apihook('WdfMemoryCreate', argc=7)
    def WdfMemoryCreate(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfMemoryCreate(
          PWDF_OBJECT_ATTRIBUTES Attributes,
          POOL_TYPE              PoolType,
          ULONG                  PoolTag,
          size_t                 BufferSize,
          WDFMEMORY              *Memory,
          PVOID                  *Buffer
        );
        '''
        DriverGlobals, Attributes, PoolType, PoolTag, BufferSize, Mem, Buf = argv

        rv = ddk.STATUS_SUCCESS

        ptr = self.mem_alloc(BufferSize, tag='api.struct.WDFMEMORY')

        if Mem:
            self.mem_write(Mem, (ptr).to_bytes(emu.get_ptr_size(), 'little'))
        if Buf:
            self.mem_write(Buf, (ptr).to_bytes(emu.get_ptr_size(), 'little'))

        return rv

    @apihook('WdfUsbTargetDeviceSelectConfig', argc=4)
    def WdfUsbTargetDeviceSelectConfig(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfUsbTargetDeviceSelectConfig(
          WDFUSBDEVICE                         UsbDevice,
          PWDF_OBJECT_ATTRIBUTES               PipeAttributes,
          PWDF_USB_DEVICE_SELECT_CONFIG_PARAMS Params
        );
        '''
        DriverGlobals, UsbDevice, PipeAttributes, Params = argv

        rv = ddk.STATUS_SUCCESS
        cfg_params = self.types.WDF_USB_DEVICE_SELECT_CONFIG_PARAMS(emu.get_ptr_size())
        cfg_params = self.mem_cast(cfg_params, Params)

        dev = self.usb_devices.get(UsbDevice)

        enum = self.types.WdfUsbTargetDeviceSelectConfigType
        if cfg_params.Type == enum.WdfUsbTargetDeviceSelectConfigTypeSingleInterface:
            dev.num_interfaces = 1
            hnd = self.get_handle()
            uf = WdfUsbInterface()
            uf.config_desc = dev.config_desc
            uf.iface_index = 0
            self.usb_interfaces.update({hnd: uf})
            cfg_params.Types.SingleInterface.ConfiguredUsbInterface = hnd
        elif cfg_params.Type == enum.WdfUsbTargetDeviceSelectConfigTypeMultiInterface:
            dev.num_interfaces = cfg_params.Types.MultiInterface.NumberInterfaces
        elif cfg_params.Type == enum.WdfUsbTargetDeviceSelectConfigTypeInterfacesDescriptor:
            dev.num_interfaces = cfg_params.Types.Descriptor.NumInterfaceDescriptors

        self.mem_write(Params, cfg_params.get_bytes())

        return rv

    @apihook('WdfUsbTargetDeviceRetrieveConfigDescriptor', argc=4)
    def WdfUsbTargetDeviceRetrieveConfigDescriptor(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfUsbTargetDeviceRetrieveConfigDescriptor(
          WDFUSBDEVICE UsbDevice,
          PVOID        ConfigDescriptor,
          PUSHORT      ConfigDescriptorLength
        );
        '''
        DriverGlobals, UsbDevice, ConfigDescriptor, ConfigDescriptorLength = argv
        rv = ddk.STATUS_BUFFER_TOO_SMALL

        buf_len = self.mem_read(ConfigDescriptorLength, 2)
        buf_len = int.from_bytes(buf_len, 'little')
        # For now, we basically assume this function is hooked since it has extremely specific
        # value
        cd = usbdefs.USB_CONFIGURATION_DESCRIPTOR()
        if ConfigDescriptor:
            cd = self.mem_cast(cd, ConfigDescriptor)
            dev = self.usb_devices.get(UsbDevice)
            if dev:
                if cd.bLength == 0:
                    self.mem_write(ConfigDescriptorLength, (cd.sizeof()).to_bytes(2,
                                                                                  'little'))
                    self.mem_write(ConfigDescriptor, cd.get_bytes())

                dev.config_desc = self.mem_read(ConfigDescriptor, buf_len)

                rv = ddk.STATUS_SUCCESS
            else:
                self.mem_write(ConfigDescriptorLength, (cd.sizeof()).to_bytes(2, 'little'))

        return rv

    @apihook('WdfUsbInterfaceSelectSetting', argc=4)
    def WdfUsbInterfaceSelectSetting(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfUsbInterfaceSelectSetting(
          WDFUSBINTERFACE                          UsbInterface,
          PWDF_OBJECT_ATTRIBUTES                   PipesAttributes,
          PWDF_USB_INTERFACE_SELECT_SETTING_PARAMS Params
        );
        '''
        DriverGlobals, UsbInterface, PipesAttributes, Params = argv

        rv = ddk.STATUS_INVALID_HANDLE
        uif = self.usb_interfaces.get(UsbInterface)
        if uif:
            ifparams = self.types.WDF_USB_INTERFACE_SELECT_SETTING_PARAMS(emu.get_ptr_size())
            ifparams = self.mem_cast(ifparams, Params)

            enum = self.types.WdfUsbTargetDeviceSelectSettingType
            if ifparams.Type == enum.WdfUsbInterfaceSelectSettingTypeSetting:
                uif.setting_index = ifparams.Types.Interface.SettingIndex
            rv = ddk.STATUS_SUCCESS

        return rv

    @apihook('WdfUsbTargetDeviceGetNumInterfaces', argc=2)
    def WdfUsbTargetDeviceGetNumInterfaces(self, emu, argv, ctx={}):
        '''
        UCHAR WdfUsbTargetDeviceGetNumInterfaces(
          WDFUSBDEVICE UsbDevice
        );
        '''
        DriverGlobals, UsbDevice = argv

        rv = 0
        dev = self.usb_devices.get(UsbDevice)
        if dev:
            rv = dev.num_interfaces

        return rv

    @apihook('WdfUsbInterfaceGetNumConfiguredPipes', argc=2)
    def WdfUsbInterfaceGetNumConfiguredPipes(self, emu, argv, ctx={}):
        '''
        BYTE WdfUsbInterfaceGetNumConfiguredPipes(
          WDFUSBINTERFACE UsbInterface
        );
        '''
        DriverGlobals, UsbInterface = argv

        rv = 0
        uif = self.usb_interfaces.get(UsbInterface)

        if uif:
            interfaces = self.parse_usb_config(uif.config_desc)
            for i, eps in interfaces:
                if uif.iface_index == i.bInterfaceNumber:
                    rv = len(eps)
                    break

        return rv

    @apihook('WdfUsbInterfaceGetNumSettings', argc=2)
    def WdfUsbInterfaceGetNumSettings(self, emu, argv, ctx={}):
        '''
        BYTE WdfUsbInterfaceGetNumSettings(
          WDFUSBINTERFACE UsbInterface
        );
        '''
        DriverGlobals, UsbInterface = argv

        rv = 0
        uif = self.usb_interfaces.get(UsbInterface)

        interfaces = self.parse_usb_config(uif.config_desc)

        for i in interfaces:
            if uif.iface_index == i[0].bInterfaceNumber:
                rv += 1

        return rv

    @apihook('WdfUsbTargetDeviceRetrieveInformation', argc=3)
    def WdfUsbTargetDeviceRetrieveInformation(self, emu, argv, ctx={}):
        '''
        NTSTATUS WdfUsbTargetDeviceRetrieveInformation(
          WDFUSBDEVICE                UsbDevice,
          PWDF_USB_DEVICE_INFORMATION Information
        );
        '''
        DriverGlobals, UsbDevice, Information = argv

        rv = ddk.STATUS_INVALID_HANDLE
        dev = self.usb_devices.get(UsbDevice)
        if dev:
            info = self.types.WDF_USB_DEVICE_INFORMATION(emu.get_ptr_size())
            info = self.mem_cast(info, Information)
            info.Size = info.sizeof()
            info.UsbdVersionInformation.USBDI_Version = 0x600
            self.mem_write(Information, info.get_bytes())
            rv = ddk.STATUS_SUCCESS

        return rv

    @apihook('WdfUsbInterfaceGetConfiguredPipe', argc=4)
    def WdfUsbInterfaceGetConfiguredPipe(self, emu, argv, ctx={}):
        '''
        WDFUSBPIPE WdfUsbInterfaceGetConfiguredPipe(
          WDFUSBINTERFACE           UsbInterface,
          UCHAR                     PipeIndex,
          PWDF_USB_PIPE_INFORMATION PipeInfo
        );
        '''
        DriverGlobals, UsbInterface, PipeIndex, PipeInfo = argv

        rv = 0
        uif = self.usb_interfaces.get(UsbInterface)

        interfaces = self.parse_usb_config(uif.config_desc)
        if PipeInfo:
            for i, eps in interfaces:
                if uif.iface_index == i.bInterfaceNumber:
                    if PipeIndex > len(eps):
                        break
                    ep = eps[PipeIndex]
                    info = self.types.WDF_USB_PIPE_INFORMATION(emu.get_ptr_size())
                    info = self.mem_cast(info, PipeInfo)
                    info.Size = info.sizeof()
                    info.MaximumPacketSize = ep.wMaxPacketSize
                    info.EndpointAddress = ep.bEndpointAddress
                    info.Interval = ep.bInterval
                    if ep.bmAttributes & usbdefs.USB_ENDPOINT_TYPE_ISOCHRONOUS:
                        info.PipeType = wdf.WdfUsbPipeTypeIsochronous
                    if ep.bmAttributes & usbdefs.USB_ENDPOINT_TYPE_BULK:
                        info.PipeType = wdf.WdfUsbPipeTypeBulk
                    if ep.bmAttributes & usbdefs.USB_ENDPOINT_TYPE_INTERRUPT:
                        info.PipeType = wdf.WdfUsbPipeTypeInterrupt
                    info.PipeType = ep.bmAttributes
                    self.mem_write(PipeInfo, info.get_bytes())

        if uif:
            hnd = self.get_handle()
            up = WdfUsbPipe()
            up.interface = uif
            up.index = PipeIndex
            self.usb_pipes.update({hnd: up})
            rv = hnd

        return rv

    @apihook('WdfUsbTargetPipeGetInformation', argc=3)
    def WdfUsbTargetPipeGetInformation(self, emu, argv, ctx={}):
        '''
        void WdfUsbTargetPipeGetInformation(
          WDFUSBPIPE                Pipe,
          PWDF_USB_PIPE_INFORMATION PipeInformation
        );
        '''
        DriverGlobals, Pipe, PipeInfo = argv

        _pipe = self.usb_pipes.get(Pipe)
        if _pipe:
            uif = _pipe.interface

            interfaces = self.parse_usb_config(uif.config_desc)

            if PipeInfo:
                for i, eps in interfaces:
                    if uif.iface_index == i.bInterfaceNumber:
                        if _pipe.index > len(eps):
                            break
                        ep = eps[_pipe.index]
                        info = self.types.WDF_USB_PIPE_INFORMATION(emu.get_ptr_size())
                        info = self.mem_cast(info, PipeInfo)
                        info.Size = info.sizeof()
                        info.MaximumPacketSize = ep.wMaxPacketSize
                        info.EndpointAddress = ep.bEndpointAddress
                        info.Interval = ep.bInterval
                        if ep.bmAttributes & usbdefs.USB_ENDPOINT_TYPE_ISOCHRONOUS:
                            info.PipeType = wdf.WDF_USB_PIPE_TYPE.WdfUsbPipeTypeIsochronous
                        if ep.bmAttributes & usbdefs.USB_ENDPOINT_TYPE_BULK:
                            info.PipeType = wdf.WDF_USB_PIPE_TYPE.WdfUsbPipeTypeBulk
                        if ep.bmAttributes & usbdefs.USB_ENDPOINT_TYPE_INTERRUPT:
                            info.PipeType = wdf.WDF_USB_PIPE_TYPE.WdfUsbPipeTypeInterrupt
                        info.PipeType = ep.bmAttributes
                        self.mem_write(PipeInfo, info.get_bytes())

        return

    @apihook('WdfUsbInterfaceGetInterfaceNumber', argc=2)
    def WdfUsbInterfaceGetInterfaceNumber(self, emu, argv, ctx={}):
        '''
        BYTE WdfUsbInterfaceGetInterfaceNumber(
          WDFUSBINTERFACE UsbInterface
        );
        '''
        DriverGlobals, UsbInterface = argv

        rv = 0
        uif = self.usb_interfaces.get(UsbInterface)

        if uif:
            rv = uif.iface_index

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import os
import json
import os.path

import unittest
import jsonschema

import speakeasy
import speakeasy.speakeasy


def get_default_config():
    fpath = os.path.join(os.path.dirname(speakeasy.__file__), 'configs', 'default.json')
    with open(fpath, 'r') as ff:
        return json.load(ff)


class TestConfig(unittest.TestCase):
    def test_speakeasy_configs(self):
        '''
        Make sure all of the bundled configs shipped with speakeasy conform
        '''
        config_dir = os.path.join(os.path.dirname(speakeasy.__file__), 'configs')
        self.assertTrue(os.path.isdir(config_dir))
        for fname in os.listdir(config_dir):
            if not fname.endswith('.json'):
                continue
            fpath = os.path.join(config_dir, fname)
            self.assertTrue(os.path.isfile(fpath))
            with open(fpath, 'r') as ff:
                config = json.load(ff)
            speakeasy.speakeasy.validate_config(config)

    def test_validation(self):
        '''
        Tweak the default config, validate that errors are caught
        '''
        conf0 = get_default_config()
        self.assertTrue('emu_engine' in conf0)
        speakeasy.speakeasy.validate_config(conf0)

        # validating a non enum
        conf0['emu_engine'] = 'alternate_engine'
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            speakeasy.speakeasy.validate_config(conf0)

        # validate a missing required field
        conf1 = get_default_config()
        conf1.pop('emu_engine', None)
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            speakeasy.speakeasy.validate_config(conf1)

        # validate an incorrect type
        conf2 = get_default_config()
        conf2['emu_engine'] = 1.0
        with self.assertRaises(jsonschema.exceptions.ValidationError):
            speakeasy.speakeasy.validate_config(conf2)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.defs.windows.com as com


class COM(object):
    """
    The Component Object Model (COM) manager for the emulator. This will manage COM interfaces.
    """
    def __init__(self, config):
        super(COM, self).__init__()
        self.interfaces = {}
        self.config = config

    def get_interface(self, name, ptr_size):
        """
        Get COM interface
        """
        iface = com.IFACE_TYPES.get(name)
        if iface:
            ci = com.ComInterface(iface, name, ptr_size)
            return ci



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import uuid
import speakeasy.winenv.defs.nt.ddk as ddk

import speakeasy.winenv.defs.wfp.fwpmtypes as fwp

from .. import api

FWP_E_NOT_FOUND = 0x80320008
FWP_E_CALLOUT_NOT_FOUND = 0x80320001
FWP_E_FILTER_NOT_FOUND = 0x80320003
FWP_E_LAYER_NOT_FOUND = 0x80320004
FWP_E_SUBLAYER_NOT_FOUND = 0x80320007


class NetIoEmuError(Exception):
    pass


class Fwpkclnt(api.ApiHandler):
    """
    Fwpkclnt is the Windows driver that implements the Windows Filtering Platform.
    For example, packet filters can be writting using this framework.
    """

    name = 'fwpkclnt'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Fwpkclnt, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.handle = 4
        self.sessions = {}
        self.injections = {}
        self.sublayers = {}
        self.callouts = {}
        self.filters = {}

        self.emu = emu

        self.fwp = fwp

        super(Fwpkclnt, self).__get_hook_attrs__(self)

    def new_id(self):
        tmp = self.handle
        self.handle += 4
        return tmp

    def new_session(self):

        # stub
        ret = self.new_id()
        self.sessions.update({ret: 'sess'})

        return ret

    def new_injection(self):
        ret = self.new_id()
        # TODO
        self.injections.update({ret: 'inj'})
        return ret

    def new_filter(self, name, desc, key):
        ret = self.new_id()
        flt = {ret: {'name': name, 'desc': desc, 'key': key}}
        self.filters.update(flt)
        return ret

    @apihook('FwpmEngineOpen0', argc=5)
    def FwpmEngineOpen0(self, emu, argv, ctx={}):
        """
        DWORD FwpmEngineOpen0(
        const wchar_t             *serverName,
        UINT32                    authnService,
        SEC_WINNT_AUTH_IDENTITY_W *authIdentity,
        const FWPM_SESSION0       *session,
        HANDLE                    *engineHandle
        );
        """

        sname, asvc, authid, sess, eng = argv

        rv = ddk.STATUS_SUCCESS

        hnd = self.new_session()

        self.mem_write(eng, hnd.to_bytes(self.get_ptr_size(), 'little'))

        return rv

    @apihook('FwpsInjectionHandleCreate0', argc=3)
    def FwpsInjectionHandleCreate0(self, emu, argv, ctx={}):
        """
        NTSTATUS FwpsInjectionHandleCreate0(
        ADDRESS_FAMILY addressFamily,
        UINT32         flags,
        HANDLE         *injectionHandle
        );
        """
        family, flags, inj_handle = argv

        rv = ddk.STATUS_SUCCESS

        hnd = self.new_injection()

        self.mem_write(inj_handle,
                       hnd.to_bytes(self.get_ptr_size(), 'little'))

        return rv

    @apihook('FwpmSubLayerAdd0', argc=3)
    def FwpmSubLayerAdd0(self, emu, argv, ctx={}):
        """
        DWORD FwpmSubLayerAdd0(
        HANDLE               engineHandle,
        const FWPM_SUBLAYER0 *subLayer,
        PSECURITY_DESCRIPTOR sd
        );
        """
        engineHandle, subLayer, sd = argv

        name = ''
        desc = ''
        weight = 0

        rv = ddk.STATUS_SUCCESS

        _sl = self.fwp.FWPM_SUBLAYER0(emu.get_ptr_size())
        layer = self.mem_cast(_sl, subLayer)
        if layer.displayData.name:
            name = self.read_wide_string(layer.displayData.name)
        if layer.displayData.description:
            desc = self.read_wide_string(layer.displayData.description)

        subkey = self.get_bytes(layer.subLayerKey)
        if all(x == 0 for x in subkey):
            subkey = uuid.uuid4().bytes
        subkey = uuid.UUID(bytes_le=subkey)
        subkey = str(subkey)

        provkey = self.get_bytes(layer.providerKey)
        if all(x == 0 for x in provkey):
            provkey = uuid.uuid4().bytes
        provkey = uuid.UUID(bytes_le=provkey)
        provkey = str(provkey)

        layer = {subkey: {'name': name, 'desc': desc, 'weight': weight,
                          'subkey': subkey, 'provkey': provkey}}

        self.sublayers.update(layer)

        return rv

    @apihook('FwpsCalloutRegister1', argc=3)
    def FwpsCalloutRegister1(self, emu, argv, ctx={}):
        """
        NTSTATUS FwpsCalloutRegister1(
          void                *deviceObject,
          const FWPS_CALLOUT1 *callout,
          UINT32              *calloutId
        );
        """
        deviceObject, pCallout, calloutId = argv

        rv = ddk.STATUS_SUCCESS
        cid = self.new_id()

        _co = self.fwp.FWPS_CALLOUT1(emu.get_ptr_size())
        callout = self.mem_cast(_co, pCallout)
        classify_fn = callout.classifyFn
        notify_fn = callout.notifyFn
        delete_fn = callout.flowDeleteFn

        co_key = self.get_bytes(callout.calloutKey)
        if all(x == 0 for x in co_key):
            co_key = uuid.uuid4().bytes
        co_key = uuid.UUID(bytes_le=co_key)
        co_key = str(co_key)

        co = {cid: {'key': co_key, 'flags': callout.flags,
                    'classify_fn': classify_fn, 'notify_fn': notify_fn,
                    'delete_fn': delete_fn}}

        self.callouts.update(co)

        if calloutId:
            self.mem_write(calloutId, cid.to_bytes(4, 'little'))

        return rv

    @apihook('FwpmCalloutAdd0', argc=4)
    def FwpmCalloutAdd0(self, emu, argv, ctx={}):
        """
        DWORD FwpmCalloutAdd0(
          HANDLE               engineHandle,
          const FWPM_CALLOUT0  *callout,
          PSECURITY_DESCRIPTOR sd,
          UINT32               *id
        );
        """
        eng, pCallout, sd, pCid = argv

        name = ''
        desc = ''
        rv = ddk.STATUS_SUCCESS

        _co = self.fwp.FWPM_CALLOUT0(emu.get_ptr_size())
        callout = self.mem_cast(_co, pCallout)
        if callout.displayData.name:
            name = self.read_wide_string(callout.displayData.name)
        if callout.displayData.description:
            desc = self.read_wide_string(callout.displayData.description)

        co_key = self.get_bytes(callout.calloutKey)
        # was the guid set to 0?
        if all(x == 0 for x in co_key):
            co_key = uuid.uuid4().bytes
        co_key = uuid.UUID(bytes_le=co_key)
        co_key = str(co_key)

        cid = 0
        for k, v in self.callouts.items():
            if v['key'] == co_key:
                cid = k
                self.callouts[k].update({'name': name, 'desc': desc})

        if pCid:
            cid = self.mem_write(pCid, cid.to_bytes(4, 'little'))

        return rv

    @apihook('FwpmFilterAdd0', argc=4)
    def FwpmFilterAdd0(self, emu, argv, ctx={}):
        """
        DWORD FwpmFilterAdd0(
          HANDLE               engineHandle,
          const FWPM_FILTER0   *filter,
          PSECURITY_DESCRIPTOR sd,
          UINT64               *id
        );
        """
        eng, pFilter, sd, pId = argv

        self.mem_write(pId, b'\x41\x41')

        rv = ddk.STATUS_SUCCESS

        name = ''
        desc = ''

        _filt = self.fwp.FWPM_FILTER0(emu.get_ptr_size())
        filt = self.mem_cast(_filt, pFilter)
        if filt.displayData.name:
            name = self.read_wide_string(filt.displayData.name)
        if filt.displayData.description:
            desc = self.read_wide_string(filt.displayData.description)

        flt_key = self.get_bytes(filt.filterKey)

        # was the guid set to 0?
        if all(x == 0 for x in flt_key):
            flt_key = uuid.uuid4().bytes

        flt_key = uuid.UUID(bytes_le=flt_key)
        flt_key = str(flt_key)

        fid = self.new_filter(name, desc, flt_key)
        self.mem_write(pId, fid.to_bytes(8, 'little'))

        return rv

    @apihook('FwpmFilterDeleteById0', argc=2)
    def FwpmFilterDeleteById0(self, emu, argv, ctx={}):
        """
        DWORD FwpmFilterDeleteById0(
        HANDLE engineHandle,
        UINT64 id
        );
        """
        eng, fid = argv

        rv = ddk.STATUS_SUCCESS

        return rv

    @apihook('FwpmCalloutDeleteById0', argc=2)
    def FwpmCalloutDeleteById0(self, emu, argv, ctx={}):
        """
        DWORD FwpmCalloutDeleteById0(
        HANDLE engineHandle,
        UINT32 id
        );
        """
        eng, cid = argv
        rv = FWP_E_CALLOUT_NOT_FOUND

        co = self.callouts.get(cid)
        if co:
            rv = ddk.STATUS_SUCCESS
        return rv

    @apihook('FwpsCalloutUnregisterById0', argc=1)
    def FwpsCalloutUnregisterById0(self, emu, argv, ctx={}):
        """
        NTSTATUS FwpsCalloutUnregisterById0(
        const UINT32 calloutId
        );
        """
        cid, = argv
        rv = FWP_E_CALLOUT_NOT_FOUND

        co = self.callouts.get(cid)
        if co:
            rv = ddk.STATUS_SUCCESS
        return rv

    @apihook('FwpmSubLayerDeleteByKey0', argc=2)
    def FwpmSubLayerDeleteByKey0(self, emu, argv, ctx={}):
        """
        DWORD FwpmSubLayerDeleteByKey0(
        HANDLE     engineHandle,
        const GUID *key
        );
        """
        eng, key = argv

        rv = FWP_E_SUBLAYER_NOT_FOUND

        guid = self.mem_read(key, 16)
        guid = uuid.UUID(bytes_le=guid)

        if self.sublayers.get(guid):
            rv = ddk.STATUS_SUCCESS

        return rv

    @apihook('FwpmEngineClose0', argc=1)
    def FwpmEngineClose0(self, emu, argv, ctx={}):
        """
        DWORD FwpmEngineClose0(
        HANDLE engineHandle
        );
        """
        eng, = argv

        rv = ddk.STATUS_SUCCESS
        return rv

    @apihook('FwpsInjectionHandleDestroy0', argc=1)
    def FwpsInjectionHandleDestroy0(self, emu, argv, ctx={}):
        """
        NTSTATUS FwpsInjectionHandleDestroy0(
        HANDLE injectionHandle
        );
        """
        handle, = argv

        rv = ddk.STATUS_SUCCESS
        if not self.injections.get(handle):
            rv = FWP_E_NOT_FOUND

        return rv



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from speakeasy.struct import EmuStruct, Ptr


class ComInterface(object):
    def __init__(self, iface, name, ptr_size):
        self.iface = iface(ptr_size)
        self.address = 0
        self.name = name


class IUnknown(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.QueryInterface = Ptr
        self.AddRef = Ptr
        self.Release = Ptr


class IMalloc(EmuStruct):
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.IUnknown = IUnknown
        self.Alloc = Ptr
        self.Realloc = Ptr
        self.Free = Ptr
        self.GetSize = Ptr
        self.DidAlloc = Ptr
        self.HeapMinimize = Ptr


IFACE_TYPES = {'IUnknown': IUnknown,
               'IMalloc':  IMalloc}



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import unittest
import ctypes as ct

from speakeasy.struct import EmuStruct, Ptr

# These are our ground truth test cases
EXPECTED_32BIT_BYTES = b'\x01\x01\x02\x02\x03\x03\x03\x03\x04\x04\x04\x04\x05\x05' \
                       b'\x06\x06\x06\x06\x07\x07\x07\x07AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA' \
                       b'\x08\x08\t\t\t\t\n\n\n\nBBBBBBBBBBBBBBBBBBBBBBBB'

EXPECTED_64BIT_BYTES = b'\x01\x01\x02\x02\x03\x03\x03\x03\x04\x04\x04\x04\x04\x04\x04\x04\x05\x05'\
                       b'\x06\x06\x06\x06\x07\x07\x07\x07AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA' \
                       b'\x08\x08\t\t\t\t\t\t\t\t\n\n\n\nBBBBBBBBBBBBBBBBBBBBBBBB'


class DEEP_NEST(EmuStruct):
    # Struct nested 2 levels deep
    def __init__(self, ptr_size):
        super().__init__(ptr_size)
        self.Field1 = ct.c_uint32
        self.Field2 = ct.c_uint32
        self.DeepData = ct.c_uint8 * 32


class SHALLOW_NEST(EmuStruct):
    # Struct nested 1 level deep
    def __init__(self, ptr_size):
        super().__init__(ptr_size, pack=1)
        self.Field1 = ct.c_uint16
        self.DeepStruct = DEEP_NEST
        self.Field2 = ct.c_uint16


class TOP_OBJECT(EmuStruct):
    # Top level struct containing nested structs
    def __init__(self, ptr_size):
        super().__init__(ptr_size=ptr_size, pack=1)
        self.Field1 = ct.c_uint16
        self.Field2 = ct.c_uint16
        self.Field3 = ct.c_uint32
        self.Field4 = Ptr  # Ptr fields allow you to dynamically create pointer-sized fields
        self.NestedStruct = SHALLOW_NEST
        self.Field5 = Ptr
        self.Field6 = ct.c_uint32
        self.Field7 = ct.c_uint8 * 24


class TestStruct(unittest.TestCase):
    def test_32bit_emit(self):
        '''
        Make sure we can convert arbitrary bytes to a speakeasy emu struct and vice versa
        '''
        # Create a TOP_OBJECT struct with a 32-bit pointer size
        ptr_size = 4
        top = TOP_OBJECT(ptr_size)

        top.Field1 = 0x0101
        top.Field2 = 0x0202
        top.Field3 = 0x03030303
        top.Field4 = 0x04040404
        top.NestedStruct.Field1 = 0x0505
        top.NestedStruct.DeepStruct.Field1 = 0x06060606
        top.NestedStruct.DeepStruct.Field2 = 0x07070707
        top.NestedStruct.DeepStruct.DeepData = b'A' * 32
        top.NestedStruct.Field2 = 0x0808
        top.Field5 = 0x09090909
        top.Field6 = 0x0a0a0a0a
        top.Field7 = b'B' * 24

        bytez = top.get_bytes()
        self.assertEqual(bytez, EXPECTED_32BIT_BYTES)

    def test_32bit_cast(self):
        '''
        Now lets flip the same test, bytes should be casted into an equally correct 32bit structure
        '''
        # Create a TOP_OBJECT struct with a 32-bit pointer size
        ptr_size = 4
        top = TOP_OBJECT(ptr_size)
        # "cast" raw bytes into a python structure for easy mutation
        top.cast(EXPECTED_32BIT_BYTES)

        # Make sure the struct fields are what we expect
        self.assertEqual(top.Field1, 0x0101)
        self.assertEqual(top.Field2, 0x0202)
        self.assertEqual(top.Field3, 0x03030303)
        self.assertEqual(top.Field4, 0x04040404)
        self.assertEqual(top.NestedStruct.Field1, 0x0505)
        self.assertEqual(top.NestedStruct.DeepStruct.Field1, 0x06060606)
        self.assertEqual(top.NestedStruct.DeepStruct.Field2, 0x07070707)
        self.assertEqual(bytes(top.NestedStruct.DeepStruct.DeepData[:]), b'A' * 32)
        self.assertEqual(top.NestedStruct.Field2, 0x0808)
        self.assertEqual(top.Field5, 0x09090909)
        self.assertEqual(top.Field6, 0x0a0a0a0a)
        self.assertEqual(bytes(top.Field7[:]), b'B' * 24)

    def test_64bit_emit(self):

        '''
        Make sure we can convert arbitrary bytes to a speakeasy emu struct and vice versa
        '''

        # Create a TOP_OBJECT struct with a 64-bit pointer size
        ptr_size = 8
        top = TOP_OBJECT(ptr_size)

        top.Field1 = 0x0101
        top.Field2 = 0x0202
        top.Field3 = 0x03030303
        top.Field4 = 0x0404040404040404
        top.NestedStruct.Field1 = 0x0505
        top.NestedStruct.DeepStruct.Field1 = 0x06060606
        top.NestedStruct.DeepStruct.Field2 = 0x07070707
        top.NestedStruct.DeepStruct.DeepData = b'A' * 32
        top.NestedStruct.Field2 = 0x0808
        top.Field5 = 0x0909090909090909
        top.Field6 = 0x0a0a0a0a
        top.Field7 = b'B' * 24

        bytez = top.get_bytes()
        self.assertEqual(bytez, EXPECTED_64BIT_BYTES)

    def test_64bit_cast(self):
        '''
        Now lets flip the same test, bytes should be casted into an equally correct 64bit structure
        '''
        # Create a TOP_OBJECT struct with a 64-bit pointer size
        ptr_size = 8
        top = TOP_OBJECT(ptr_size)

        # "cast" raw bytes into a python structure for easy mutation
        top.cast(EXPECTED_64BIT_BYTES)

        # Make sure the struct fields are what we expect
        self.assertEqual(top.Field1, 0x0101)
        self.assertEqual(top.Field2, 0x0202)
        self.assertEqual(top.Field3, 0x03030303)
        self.assertEqual(top.Field4, 0x0404040404040404)
        self.assertEqual(top.NestedStruct.Field1, 0x0505)
        self.assertEqual(top.NestedStruct.DeepStruct.Field1, 0x06060606)
        self.assertEqual(top.NestedStruct.DeepStruct.Field2, 0x07070707)
        self.assertEqual(bytes(top.NestedStruct.DeepStruct.DeepData[:]), b'A' * 32)
        self.assertEqual(top.NestedStruct.Field2, 0x0808)
        self.assertEqual(top.Field5, 0x0909090909090909)
        self.assertEqual(top.Field6, 0x0a0a0a0a)
        self.assertEqual(bytes(top.Field7[:]), b'B' * 24)


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch
from speakeasy.errors import ApiEmuError
from speakeasy.winenv.api.kernelmode import ntoskrnl, hal, wdfldr, netio, ndis, fwpkclnt, usbd
from speakeasy.winenv.api.usermode import ws2_32, kernel32, wininet, winhttp, user32, \
                                          advapi32, msvcrt, wtsapi32, mscoree, dnsapi, \
                                          ntdll, crypt32, shell32, shlwapi, advpack, gdi32, \
                                          urlmon, ole32, comctl32, msimg32, msi32, lz32

API_HANDLERS = (
                    # Kernel mode
                    ('ntoskrnl', ntoskrnl.Ntoskrnl),
                    ('wdfldr', wdfldr.Wdfldr),
                    ('hal', hal.Hal),
                    ('usbd', usbd.Usbd),
                    ('netio', netio.Netio),
                    ('ndis', ndis.Ndis),
                    ('fwpkclnt', fwpkclnt.Fwpkclnt),
                    # User mode
                    ('ws2_32', ws2_32.Ws2_32),
                    ('kernel32', kernel32.Kernel32),
                    ('ntdll', ntdll.Ntdll),
                    ('wininet', wininet.Wininet),
                    ('winhttp', winhttp.WinHttp),
                    ('user32', user32.User32),
                    ('msvcrt', msvcrt.Msvcrt),
                    ('wtsapi32', wtsapi32.WtsApi32),
                    ('advapi32', advapi32.AdvApi32),
                    ('dnsapi', dnsapi.DnsApi),
                    ('mscoree', mscoree.Mscoree),
                    ('crypt32', crypt32.Crypt32),
                    ('shell32', shell32.Shell32),
                    ('shlwapi', shlwapi.Shlwapi),
                    ('advpack', advpack.Advpack),
                    ('gdi32', gdi32.GDI32),
                    ('urlmon', urlmon.Urlmon),
                    ('ole32', ole32.Ole32),
                    ('comctl32', comctl32.Comctl32),
                    ('msimg32', msimg32.Msimg32),
                    ('msi', msi32.Msi32),
                    ('lz32', lz32.Lz32),
               )


class WindowsApi:

    def __init__(self, emu):
        self.mods = {}
        self.instances = []
        self.data = {}
        self.emu = emu
        arch = self.emu.get_arch()

        if arch == _arch.ARCH_X86:
            self.ptr_size = 4
        elif arch == _arch.ARCH_AMD64:
            self.ptr_size = 8
        else:
            raise ApiEmuError('Invalid architecture')

    def load_api_handler(self, mod_name):
        for name, hdl in API_HANDLERS:
            name = name.lower()
            if mod_name and name == mod_name.lower():
                handler = self.mods.get(name)
                if not handler:
                    handler = hdl(self.emu)
                    self.mods.update({name: handler})
                return handler
        return None

    def get_data_export_handler(self, mod_name, exp_name):
        mod = self.mods.get(mod_name)
        if not mod:
            mod = self.load_api_handler(mod_name)
        if not mod:
            return None, None
        return (mod, mod.get_data_handler(exp_name))

    def get_export_func_handler(self, mod_name, exp_name):
        mod = self.mods.get(mod_name)
        if not mod:
            mod = self.load_api_handler(mod_name)
        if not mod:
            return None, None
        return (mod, mod.get_func_handler(exp_name))

    def call_api_func(self, mod, func, argv, ctx):
        """
        Call the handler to implement the imported API
        """
        return func(mod, self.emu, argv, ctx)

    def call_data_func(self, mod, func, ptr):
        """
        Call the handler to initialize and return imported data variables
        """
        return func(mod, ptr)



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

from .. import api


class GDI32(api.ApiHandler):

    """
    Implements exported functions from gdi32.dll
    """

    name = 'gdi32'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(GDI32, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.handle = 0
        self.count = 0
        super(GDI32, self).__get_hook_attrs__(self)

    def get_handle(self):
        self.handle += 4
        hnd = self.handle
        return hnd

    @apihook('CreateBitmap', argc=5)
    def CreateBitmap(self, emu, argv, ctx={}):
        '''
        HBITMAP CreateBitmap(
            int        nWidth,
            int        nHeight,
            UINT       nPlanes,
            UINT       nBitCount,
            const VOID *lpBits
        );
        '''
        return self.get_handle()

    @apihook('MoveToEx', argc=1)
    def MoveToEx(self, emu, argv, ctx={}):
        """
        BOOL MoveToEx(
          HDC     hdc,
          int     x,
          int     y,
          LPPOINT lppt
        );
        """
        return 1

    @apihook('LineTo', argc=1)
    def LineTo(self, emu, argv, ctx={}):
        """
        BOOL LineTo(
          HDC hdc,
          int x,
          int y
        )
        """
        return 1

    @apihook('GetStockObject', argc=1)
    def GetStockObject(self, emu, argv, ctx={}):
        """
        HGDIOBJ GetStockObject(
            int i
        );
        """
        return 0

    @apihook('GetMapMode', argc=1)
    def GetMapMode(self, emu, argv, ctx={}):
        """
        int GetMapMode(
            HDC hdc
        );
        """
        return 1

    @apihook('GetDeviceCaps', argc=2)
    def GetDeviceCaps(self, emu, argv, ctx={}):
        """
        int GetDeviceCaps(
            HDC hdc,
            int index
        );
        """
        return 16

    @apihook('GdiSetBatchLimit', argc=1)
    def GdiSetBatchLimit(self, emu, argv, ctx={}):
        """
        DWORD GdiSetBatchLimit(
          DWORD dw
        );
        """
        return 0

    @apihook('MaskBlt', argc=12)
    def MaskBlt(self, emu, argv, ctx={}):
        """
        BOOL MaskBlt(
          HDC     hdcDest,
          int     xDest,
          int     yDest,
          int     width,
          int     height,
          HDC     hdcSrc,
          int     xSrc,
          int     ySrc,
          HBITMAP hbmMask,
          int     xMask,
          int     yMask,
          DWORD   rop
        );
        """
        return 1

    @apihook('SelectObject', argc=2)
    def SelectObject(self, emu, argv, ctx={}):
        """
        HGDIOBJ SelectObject(
          HDC     hdc,
          HGDIOBJ h
        );
        """
        return 0

    @apihook('GetTextCharacterExtra', argc=1)
    def GetTextCharacterExtra(self, emu, argv, ctx={}):
        """
        int GetTextCharacterExtra(
          HDC hdc
        );
        """
        return 0x8000000

    @apihook('StretchBlt', argc=11)
    def StretchBlt(self, emu, argv, ctx={}):
        """
        BOOL StretchBlt(
          HDC   hdcDest,
          int   xDest,
          int   yDest,
          int   wDest,
          int   hDest,
          HDC   hdcSrc,
          int   xSrc,
          int   ySrc,
          int   wSrc,
          int   hSrc,
          DWORD rop
        );
        """
        return 0



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

class SpeakeasyError(Exception):
    """
    Base class for Speakeasy errors
    """
    pass


class NotSupportedError(SpeakeasyError):
    """
    Sample is not currently supported
    """
    pass


class ApiEmuError(SpeakeasyError):
    """
    Base class for API errors
    """
    pass


class EmuException(SpeakeasyError):
    """
    Base class for emulation errors
    """
    pass


class EmuEngineError(SpeakeasyError):
    pass


class WindowsEmuError(SpeakeasyError):
    """
    Base class for Windows emulation errors
    """
    pass


class KernelEmuError(SpeakeasyError):
    """
    Base class for Windows kernel mode emulation errors
    """
    pass


class Win32EmuError(SpeakeasyError):
    """
    Base class for Windows user mode emulation errors
    """
    pass


class FileSystemEmuError(SpeakeasyError):
    """
    Base class for Windows user mode emulation errors
    """
    pass


class NetworkEmuError(SpeakeasyError):
    """
    Raised during network emulation errors
    """
    pass


class RegistryEmuError(SpeakeasyError):
    """
    Raised during registry emulation errors
    """
    pass


class ConfigError(SpeakeasyError):
    """
    Raised during validating configuration
    """
    pass



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.arch as _arch
from speakeasy.profiler import Run
from speakeasy.struct import EmuStruct
from speakeasy.errors import ApiEmuError
import speakeasy.windows.common as winemu

import speakeasy.winenv.defs.nt.ntoskrnl as ntos


class ApiHandler(object):
    """
    Base class for handling exported functions
    """

    name = ''

    @staticmethod
    def apihook(impname, argc=0, conv=_arch.CALL_CONV_STDCALL, ordinal=None):

        def apitemp(f):
            if not callable(f):
                raise ApiEmuError('Invalid function type supplied: %s' % (str(f)))
            f.__apihook__ = (impname, f, argc, conv, ordinal)
            return f

        return apitemp

    @staticmethod
    def impdata(impname):

        def datatmp(f):
            if not callable(f):
                raise ApiEmuError('Invalid function type supplied: %s' % (str(f)))
            f.__datahook__ = (impname, f)
            return f

        return datatmp

    @staticmethod
    def get_api_name(func):
        return func.__apihook__[0]

    def __init__(self, emu):
        super(ApiHandler, self).__init__()
        self.funcs = {}
        self.data = {}
        self.mod_name = ''
        self.emu = emu
        arch = self.emu.get_arch()

        if arch == _arch.ARCH_X86:
            self.ptr_size = 4
        elif arch == _arch.ARCH_AMD64:
            self.ptr_size = 8
        else:
            raise ApiEmuError('Invalid architecture')

        for name in dir(self):
            val = getattr(self, name, None)
            if val is None:
                continue

            func_attrs = getattr(val, '__apihook__', None)
            data_attrs = getattr(val, '__datahook__', None)
            if func_attrs:
                name, func, argc, conv, ordinal = func_attrs
                self.funcs[name] = (name, func, argc, conv, ordinal)
                if ordinal:
                    self.funcs[ordinal] = (name, func, argc, conv, ordinal)

            elif data_attrs:
                name, func = data_attrs
                self.data[name] = func

    def __get_hook_attrs__(self, obj):
        for name in dir(obj):
            val = getattr(obj, name, None)
            if val is None:
                continue

            func_attrs = getattr(val, '__apihook__', None)
            data_attrs = getattr(val, '__datahook__', None)
            if func_attrs:
                name, func, argc, conv, ordinal = func_attrs
                obj.funcs[name] = (name, func, argc, conv, ordinal)
                if ordinal:
                    obj.funcs[ordinal] = (name, func, argc, conv, ordinal)

            elif data_attrs:
                name, func = data_attrs
                obj.data[name] = func

    def get_data_handler(self, exp_name):
        return self.data.get(exp_name)

    def get_func_handler(self, exp_name):
        if exp_name.startswith('ordinal_'):
            ord_num = exp_name.split('_')
            if len(ord_num) == 2 and ord_num[1].isdigit():
                ord_num = int(ord_num[1])
                handler = self.funcs.get(ord_num)
                if handler:
                    return handler
        return self.funcs.get(exp_name)

    def get_ptr_size(self):
        return self.ptr_size

    def sizeof(self, obj):
        if isinstance(obj, EmuStruct):
            return obj.sizeof()
        else:
            raise ApiEmuError('Invalid object')

    def get_bytes(self, obj):
        if isinstance(obj, EmuStruct):
            return obj.get_bytes()
        else:
            raise ApiEmuError('Invalid object')

    def cast(self, obj, bytez):
        if isinstance(obj, EmuStruct):
            return obj.cast(bytez)
        else:
            raise ApiEmuError('Invalid object')
        return obj

    def write_back(self, addr, obj):
        bytez = self.get_bytes(obj)
        self.emu.mem_write(addr, bytez)

    def pool_alloc(self, pool_type, size, tag):
        return self.emu.pool_alloc(pool_type, size, tag)

    def heap_alloc(self, size, heap):
        return self.emu.heap_alloc(size, heap)

    def mem_alloc(self, size, base=None, tag=None, flags=0, perms=0, shared=False, process=None):
        return self.emu.mem_map(size, base=base, tag=tag, flags=flags, perms=perms,
                                shared=shared, process=process)

    def mem_free(self, addr):
        return self.emu.mem_free(addr)

    def mem_reserve(self, size, base=None, tag=None):
        return self.emu.mem_reserve(size, base=base, tag=tag)

    def mem_cast(self, obj, addr):
        struct_bytes = self.emu.mem_read(addr, self.sizeof(obj))
        return self.cast(obj, struct_bytes)

    def mem_copy(self, dst, src, n):
        return self.emu.mem_copy(dst, src, n)

    def read_mem_string(self, addr, width, max_chars=0):
        string = self.emu.read_mem_string(addr, width=width)
        return string

    def mem_string_len(self, addr, width):
        return self.emu.mem_string_len(addr, width)

    def read_ansi_string(self, addr):
        ans = ntos.STRING(self.emu.get_ptr_size())
        ans = self.mem_cast(ans, addr)

        string = self.emu.read_mem_string(ans.Buffer, width=1)
        return string

    def read_unicode_string(self, addr):
        us = ntos.UNICODE_STRING(self.emu.get_ptr_size())
        us = self.mem_cast(us, addr)

        string = self.emu.read_mem_string(us.Buffer, width=2)
        return string

    def read_wide_string(self, addr, max_chars=0):
        string = self.emu.read_mem_string(addr, width=2, max_chars=max_chars)
        return string

    def read_string(self, addr, max_chars=0):
        string = self.emu.read_mem_string(addr, width=1, max_chars=max_chars)
        return string

    def write_mem_string(self, string, addr, width):
        return self.emu.write_mem_string(string, addr, width)

    def write_wide_string(self, string, addr):
        return self.write_mem_string(string, addr, width=2)

    def write_string(self, string, addr):
        return self.write_mem_string(string, addr, width=1)

    def queue_run(self, run_type, ep, run_args=[]):
        run = Run()
        if not isinstance(run_type, str):
            raise ApiEmuError('Invalid run type')
        if not isinstance(ep, int):
            raise ApiEmuError('Invalid run entry point')
        if not any((isinstance(run_args, list), isinstance(run_args, tuple))):
            raise ApiEmuError('Invalid run args')

        run.type = run_type
        run.start_addr = ep
        run.args = run_args
        self.emu.add_run(run)

    def log_file_access(self, path, event_type, data=None,
                        handle=0, disposition=[], access=[], buffer=0,
                        size=None):
        profiler = self.emu.get_profiler()
        if profiler:
            run = self.emu.get_current_run()
            profiler.log_file_access(run, path, event_type, data, handle,
                                     disposition, access, buffer, size)

    def log_process_event(self, proc, event_type, **kwargs):
        profiler = self.emu.get_profiler()
        if profiler:
            run = self.emu.get_current_run()
            profiler.log_process_event(run, proc, event_type, kwargs)

    def log_registry_access(self, path, event_type, value_name=None, data=None,
                            handle=0, disposition=[], access=[], buffer=0,
                            size=None):
        profiler = self.emu.get_profiler()
        if profiler:
            run = self.emu.get_current_run()
            profiler.log_registry_access(run, path, event_type, value_name, data, handle,
                                         disposition, access, buffer, size)

    def log_dns(self, domain, ip=''):
        profiler = self.emu.get_profiler()
        if profiler:
            run = self.emu.get_current_run()
            profiler.log_dns(run, domain, ip)

    def log_network(self, server, port, typ='unknown', proto='unknown', data=b'', method=''):
        profiler = self.emu.get_profiler()
        if profiler:
            run = self.emu.get_current_run()
            profiler.log_network(run, server, port, typ=typ, proto=proto,
                                 data=data, method=method)

    def log_http(self, server, port, headers='', body=b'', secure=False):
        profiler = self.emu.get_profiler()
        if profiler:
            run = self.emu.get_current_run()
            profiler.log_http(run, server, port, headers=headers,
                              body=body, secure=secure)

    def get_max_int(self):
        # Byte order is irrelevant here
        return int.from_bytes(b'\xFF' * self.get_ptr_size(), 'little')

    def mem_read(self, addr, size):
        return self.emu.mem_read(addr, size)

    def file_open(self, path, create=False):
        return self.emu.file_open(path, create)

    def file_create_mapping(self, hfile, name, size, prot):
        return self.emu.file_create_mapping(hfile, name, size, prot)

    def file_get(self, handle):
        return self.emu.file_get(handle)

    def does_file_exist(self, path):
        return self.emu.does_file_exist(path)

    def reg_open_key(self, path, create=False):
        return self.emu.reg_open_key(path, create)

    def reg_get_key(self, handle):
        return self.emu.reg_get_key(handle)

    def reg_get_subkeys(self, hkey):
        return self.emu.reg_get_subkeys(hkey)

    def get_encoding(self, char_width):
        if char_width == 2:
            enc = 'utf-16le'
        elif char_width == 1:
            enc = 'utf-8'
        else:
            raise ApiEmuError('No encoding found for char width: %d' % (char_width))
        return enc

    def mem_write(self, addr, data):

        # If the data being written to a shared memory mapping, update all mappings
        # This will likely have to be made more robust to handle more complicated
        # scenarios with varying file offsets
        mm = self.emu.get_address_map(addr)
        if mm and mm.shared:
            fm = self.emu.get_file_manager()
            fmap = fm.get_mapping_from_addr(mm.get_base())
            if fmap:
                for base, view in fmap.views.items():
                    if base == mm.get_base():
                        continue
                    tgt_offset = addr - mm.get_base()
                    self.emu.mem_write(base + tgt_offset, data)

        return self.emu.mem_write(addr, data)

    def create_thread(self, addr, ctx, hproc, thread_type='thread', is_suspended=False):
        return self.emu.create_thread(addr, ctx, hproc, thread_type=thread_type,
                                      is_suspended=is_suspended)

    def get_object_from_id(self, id):
        return self.emu.get_object_from_id(id)

    def get_object_from_addr(self, addr):
        return self.emu.get_object_from_addr(addr)

    def get_object_handle(self, obj):
        return self.emu.get_object_handle(obj)

    def get_object_from_handle(self, hnd):
        return self.emu.get_object_from_handle(hnd)

    def get_object_from_name(self, name):
        return self.emu.get_object_from_name(name)

    def get_os_version(self):
        return self.emu.osversion

    def exit_process(self):
        self.emu.exit_process()

    def get_char_width(self, ctx):
        """
        Based on the API name, determine the character width
        being used by the function
        """
        name = ctx.get('func_name', '')
        if name.endswith('A'):
            return 1
        elif name.endswith('W'):
            return 2
        raise ApiEmuError('Failed to get character width from function: %s' % (name))

    def get_va_arg_count(self, fmt):
        """
        Get the number of arguments in the variable argument list
        """

        # Ignore escapes
        i = fmt.count('%%')
        c = fmt.count('%')

        if self.get_ptr_size() != 8:
            c += fmt.count('%ll')
        return c - i

    def va_args(self, va_list, num_args):
        """
        Get the variable argument list
        """
        args = []
        ptr = va_list
        ptrsize = self.get_ptr_size()

        for n in range(num_args):
            arg = int.from_bytes(self.emu.mem_read(ptr, ptrsize), 'little')
            args.append(arg)
            ptr += ptrsize
        return args

    def setup_callback(self, func, args, caller_argv=[]):
        """
        For APIs that call functions, we will setup the stack to make this flow
        naturally.
        """

        run = self.emu.get_current_run()

        if not len(run.api_callbacks):
            # Get the original return address
            ret = self.emu.get_ret_address()
            sp = self.emu.get_stack_ptr()

            self.emu.set_func_args(sp, winemu.API_CALLBACK_HANDLER_ADDR, *args)
            self.emu.set_pc(func)
            run.api_callbacks.append((ret, func, caller_argv))
        else:
            run.api_callbacks.append((None, func, args))

    def do_str_format(self, string, argv):
        """
        Format a string similar to msvcrt.printf
        """

        # Skip over the format string
        args = list(argv)
        new = list(string)
        curr_fmt = ''
        new_fmts = []

        # Very brittle format string parser, should improve later
        inside_fmt = False
        for i, c in enumerate(string):

            if c == '%':
                if inside_fmt:
                    inside_fmt = False
                else:
                    inside_fmt = True

            if inside_fmt:
                if c == 'S':
                    s = self.read_wide_string(args.pop(0))
                    new_fmts.append(s)
                    new[i] = 's'
                    inside_fmt = False

                elif c == 's':
                    if curr_fmt.startswith('w'):
                        s = self.read_wide_string(args.pop(0))
                        new[i - 1] = '\xFF'
                        curr_fmt = ''
                        new_fmts.append(s)
                    else:
                        s = self.read_string(args.pop(0))
                        new_fmts.append(s)
                elif c in ('x', 'X', 'd', 'u', 'i'):
                    if curr_fmt.startswith('ll'):
                        if self.get_ptr_size() == 8:
                            new_fmts.append(args.pop(0))
                        else:
                            low = args.pop(0)
                            high = args.pop(0)
                            new_fmts.append(high << 32 | low)
                        new = new[: i - 2] + new[i:]
                        curr_fmt = ''
                    else:
                        new_fmts.append(0xFFFFFFFF & args.pop(0))
                elif c == 'c':
                    new_fmts.append(0xFF & args.pop(0))
                elif c == 'P':
                    new[i] = 'X'
                    new_fmts.append(args.pop(0))
                elif c == 'p':
                    new[i] = 'x'
                    new_fmts.append(args.pop(0))
                elif c == 'l':
                    curr_fmt += c
                elif c == 'w':
                    curr_fmt += c

            if inside_fmt and c in 'diuoxXfFeEgGaAcspn':
                inside_fmt = False

            if not args:
                break

        new = ''.join(new)
        new = new.replace('\xFF', '')
        new = new % tuple(new_fmts)

        return new



# Copyright (C) 2020 FireEye, Inc. All Rights Reserved.

import speakeasy.winenv.defs.nt.ddk as ddk

import speakeasy.winenv.defs.nt.ntoskrnl as nt
import speakeasy.winenv.defs.ndis.ndis as ndis

from .. import api


NDIS_STATUS_FAILURE = 0xC0000001
NDIS_STATUS_SUCCESS = 0x00000000
NDIS_STATUS_RESOURCES = 0xC000009A


class Ndis(api.ApiHandler):
    """
    Implements the Network Driver Interface Specification (NDIS) used for
    interacting with network cards.
    """

    name = 'ndis'
    apihook = api.ApiHandler.apihook
    impdata = api.ApiHandler.impdata

    def __init__(self, emu):

        super(Ndis, self).__init__(emu)

        self.funcs = {}
        self.data = {}
        self.handle = 4
        self.wrappers = {}
        self.drivers = {}
        self.protocols = {}

        self.emu = emu

        self.win = nt
        self.ndis = ndis
        super(Ndis, self).__get_hook_attrs__(self)

    def convert_pool_tag(self, tag):
        ret = '0x00'
        if tag:
            try:
                ret = tag.to_bytes(4, 'little').decode('utf-8')
            except Exception:
                ret = '0x%x' % tag
        return ret

    def new_id(self):
        tmp = self.handle
        self.handle += 4
        return tmp

    @apihook('NdisGetVersion', argc=0)
    def NdisGetVersion(self, emu, argv, ctx={}):
        """
        UINT NdisGetVersion();
        """

        ndis_major = 5
        ndis_minor = 0
        osver = self.get_os_version()
        major, minor = osver['major'], osver['minor']

        if major >= 6:
            ndis_major = 6

        if major >= 10:
            ndis_minor = 0x51
        elif minor >= 1:
            ndis_minor = 1

        out_ver = (ndis_major << 16) | ndis_minor

        return out_ver

    @apihook('NdisGetRoutineAddress', argc=1)
    def NdisGetRoutineAddress(self, emu, argv, ctx={}):
        """
        PVOID NdisGetRoutineAddress(
            PNDIS_STRING NdisRoutineName
        );
        """
        NdisRoutineName, = argv
        fn = self.read_unicode_string(NdisRoutineName)

        addr = emu.get_proc('ndis', fn)
        argv[0] = fn
        return addr

    @apihook('NdisMRegisterMiniportDriver', argc=5)
    def NdisMRegisterMiniportDriver(self, emu, argv, ctx={}):
        """
        NDIS_STATUS NdisMRegisterMiniportDriver(
            PDRIVER_OBJECT DriverObject,
            PUNICODE_STRING RegistryPath,
            NDIS_HANDLE MiniportDriverContext,
            PNDIS_MINIPORT_DRIVER_CHARACTERISTICS MiniportDriverCharacteristics,
            PNDIS_HANDLE NdisMiniportDriverHandle
        );
        """
        drv, reg, drv_ctx, chars, phnd = argv
        rv = NDIS_STATUS_SUCCESS

        if phnd:
            hnd = self.new_id()
            hnd = hnd.to_bytes(4, 'little')
            self.mem_write(phnd, hnd)
        return rv

    @apihook('NdisInitializeWrapper', argc=4)
    def NdisInitializeWrapper(self, emu, argv, ctx={}):
        """
        VOID NdisInitializeWrapper(
            PNDIS_HANDLE    NdisWrapperHandle,
            PVOID           SystemSpecific1,
            PVOID           SystemSpecific2,
            PVOID           SystemSpecific3)
        """
        pHandle, ss1, ss2, ss3 = argv

        hnd = self.new_id()
        self.wrappers.update({hnd: True})
        self.mem_write(pHandle, hnd.to_bytes(self.get_ptr_size(), 'little'))

    @apihook('NdisTerminateWrapper', argc=2)
    def NdisTerminateWrapper(self, emu, argv, ctx={}):
        """
        VOID NdisTerminateWrapper(
        _In_ NDIS_HANDLE NdisWrapperHandle,
        _In_ PVOID       SystemSpecific
        );
        """
        hnd, ss = argv

    @apihook('NdisInitializeReadWriteLock', argc=1)
    def NdisInitializeReadWriteLock(self, emu, argv, ctx={}):
        """
        void NdisInitializeReadWriteLock(
            PNDIS_RW_LOCK Lock
        );
        """
        lock, = argv

    @apihook('NdisMRegisterUnloadHandler', argc=2)
    def NdisMRegisterUnloadHandler(self, emu, argv, ctx={}):
        """
        VOID NdisMRegisterUnloadHandler(
        _In_ NDIS_HANDLE    NdisWrapperHandle,
        _In_ PDRIVER_UNLOAD UnloadHandler
        );
        """
        hnd, unload = argv

    @apihook('NdisRegisterProtocol', argc=4)
    def NdisRegisterProtocol(self, emu, argv, ctx={}):
        """
        VOID NdisRegisterProtocol(
        _Out_ PNDIS_STATUS                   Status,
        _Out_ PNDIS_HANDLE                   NdisProtocolHandle,
        _In_  PNDIS_PROTOCOL_CHARACTERISTICS ProtocolCharacteristics,
        _In_  UINT                           CharacteristicsLength
        );
        """
        pStatus, pProtoHandle, pChars, clen = argv
        rv = NDIS_STATUS_SUCCESS
        hnd = self.new_id()
        pchr = self.mem_read(pChars, clen)

        self.protocols.update({hnd: pchr})

        if pStatus:
            self.mem_write(pStatus, rv.to_bytes(4, 'little'))

        if pProtoHandle:
            self.mem_write(pProtoHandle, hnd.to_bytes(4, 'little'))

    @apihook('NdisIMRegisterLayeredMiniport', argc=4)
    def NdisIMRegisterLayeredMiniport(self, emu, argv, ctx={}):
        """
        NDIS_STATUS NdisIMRegisterLayeredMiniport(
        _In_  NDIS_HANDLE                    NdisWrapperHandle,
        _In_  PNDIS_MINIPORT_CHARACTERISTICS MiniportCharacteristics,
        _In_  UINT                           CharacteristicsLength,
        _Out_ PNDIS_HANDLE                   DriverHandle
        );
        """
        hnd, mp_chars, clen, drv_hnd = argv
        rv = NDIS_STATUS_SUCCESS

        if not self.wrappers.get(hnd):
            rv = NDIS_STATUS_FAILURE

        dhnd = self.new_id()
        mpc = self.mem_read(mp_chars, clen)
        self.drivers.update({dhnd: mpc})

        if drv_hnd:
            self.mem_write(drv_hnd, dhnd.to_bytes(4, 'little'))

        return rv

    @apihook('NdisIMAssociateMiniport', argc=2)
    def NdisIMAssociateMiniport(self, emu, argv, ctx={}):
        """
        void NdisIMAssociateMiniport(
        NDIS_HANDLE DriverHandle,
        NDIS_HANDLE ProtocolHandle
        );
        """
        drv_hnd, phnd = argv

    @apihook('NdisAllocateGenericObject', argc=3)
    def NdisAllocateGenericObject(self, emu, argv, ctx={}):
        """
        PNDIS_GENERIC_OBJECT NdisAllocateGenericObject(
            PDRIVER_OBJECT DriverObject,
            ULONG          Tag,
            USHORT         Size
        );
        """
        drv, tag, size = argv

        ptr = 0

        stag = self.convert_pool_tag(tag)
        argv[1] = stag

        go = self.ndis.NDIS_GENERIC_OBJECT(emu.get_ptr_size())
        go.DriverObject = drv

        total = size + self.sizeof(go)
        ptr = self.mem_alloc(size=total,
                             tag='api.struct.NDIS_GENERIC_OBJECT.%s' % (stag))
        self.mem_write(ptr, self.get_bytes(go))

        return ptr

    @apihook('NdisAllocateMemoryWithTag', argc=3)
    def NdisAllocateMemoryWithTag(self, emu, argv, ctx={}):
        """
        NDIS_STATUS NdisAllocateMemoryWithTag(
          _Out_ PVOID *VirtualAddress,
          _In_  UINT  Length,
          _In_  ULONG Tag
        );
        """
        va, size, tag = argv

        rv = ddk.STATUS_SUCCESS

        stag = self.convert_pool_tag(tag)
        argv[2] = stag

        ptr = self.mem_alloc(size=size, tag='api.ndis_pool.%s' % (stag))
        self.mem_write(va,
                       ptr.to_bytes(emu.get_ptr_size(), 'little'))

        return rv

    @apihook('NdisAllocateNetBufferListPool', argc=2)
    def NdisAllocateNetBufferListPool(self, emu, argv, ctx={}):
        """
        NDIS_HANDLE NdisAllocateNetBufferListPool(
          NDIS_HANDLE                      NdisHandle,
          PNET_BUFFER_LIST_POOL_PARAMETERS Parameters
        );
        """
        NdisHandle, Parameters = argv

        params = self.mem_cast(self.ndis.NET_BUFFER_LIST_POOL_PARAMETERS(emu.get_ptr_size()),
                               Parameters)

        nbl = self.ndis.NET_BUFFER_LIST(emu.get_ptr_size())

        if params.fAllocateNetBuffer:
            net_buf = self.ndis.NET_BUFFER(emu.get_ptr_size())
            nb_ptr = self.mem_alloc(size=self.sizeof(net_buf),
                                    tag='api.struct.NET_BUFFER')
            nbl.FirstNetBuffer = nb_ptr

            self.mem_write(nb_ptr, self.get_bytes(net_buf))

        nbl_ptr = self.mem_alloc(size=self.sizeof(nbl),
                                 tag='api.struct.NET_BUFFER_LIST')
        self.mem_write(nbl_ptr, self.get_bytes(nbl))

        return nbl_ptr

    @apihook('NdisFreeNetBufferListPool', argc=1)
    def NdisFreeNetBufferListPool(self, emu, argv, ctx={}):
        """
        void NdisFreeNetBufferListPool(
        NDIS_HANDLE PoolHandle
        );
        """
        handle, = argv

        return

    @apihook('NdisFreeMemory', argc=3)
    def NdisFreeMemory(self, emu, argv, ctx={}):
        """
        void NdisFreeMemory(
        PVOID VirtualAddress,
        UINT  Length,
        UINT  MemoryFlags
        );
        """
        va, length, flags = argv

        return

    @apihook('NdisFreeGenericObject', argc=1)
    def NdisFreeGenericObject(self, emu, argv, ctx={}):
        """
        void NdisFreeGenericObject(
        PNDIS_GENERIC_OBJECT NdisObject
        );
        """
        pObj,  = argv

        return
